#include "complexmath.h"
#include <cmath>	// in w2k3ddk/inc/crt/

// Unregelmigkeiten in <cmath> bgeln; hypot() und cabs() machen dasselbe
#ifdef _M_X64
// hypot() gibt's bei 64 Bit nicht als berladene Funktion
static float hypot(float x,float y) {return _hypotf(x,y);}
#else
// hypot() gibt's bei 32 Bit nicht als Float-Version, jedenfalls nicht in der msvcrt-light.lib
static float hypot(float x,float y) {return float(_hypot(x,y));}
#endif

/*							*
 * complex svw. std::complex, nur ohne Exceptionhandler	*
 *							*/

// 2 komplexe Zahlen multiplizieren
template<class T>
complex<T> complex<T>::operator*(const complex<T>&o) const{
 return complex(x*o.x-y*o.y,x*o.y+y*o.x);
}

// 2 komplexe Zahlen dividieren
template<class T>
complex<T> complex<T>::operator/(const complex<T>&o) const{
 T n = o.x*o.x+o.y*o.y;	// das Quadrat des komplexen Betrags
 return complex((x*o.x+y*o.y)/n,(y*o.x-x*o.y)/n);
}

template<class T> T radians2angle(T x,angle_t ang) {
 if (ang) x*=int(ang)/T(PI2); // radians => turns, degrees, gradians
 return x;
}

template<class T> T radians_norm(T x) {	// good performance only for small x
 while (x>=T(PI_2)) x-=T(PI);
 while (x<-T(PI_2)) x+=T(PI);
 return x;
}

template<class T>
T complex<T>::arg(angle_t ang) const {
 return radians2angle<T>(atan2(y,x),ang);
}

template<class T>
T complex<T>::abs() const {return hypot(x,y);}

template<class T>
complex<T>& complex<T>::set(T abs, T arg, angle_t ang) {
 if (ang) arg*=T(PI2)/int(ang);	// turns, degrees, gradians => radians
 return set(abs*cos(arg),abs*sin(arg));	// optimizer may invoke fsincos 80x87 instruction
}

template<class T>
complex<T> complex<T>::operator^(char exp) const{
 complex<T> r;
 if (exp>=0) return r.set(::pow(abs(),exp),exp*arg(radians),radians);
 T n=x*x+y*y;
 return r.set(x/n,-y/n);	// Invers fr exp=-1 (hier fr alle negativen Zahlen)
}

template<class T> IqData<T> DFT(const IqData<T>&data) {
 unsigned N = static_cast<unsigned>(data.size());
 IqData<T> ret(N);			// Zieldaten (complex<> initialisiert mit 0)
 for (unsigned i=0; i<N; i++) {		// loop over all target values
  complex<T> wActual(1,0);		// actual rotation factor
  complex<T> wStep(1,-T(i)/N,turns);	// step between rotation factors
  for (unsigned j=0; j<N; j++) {	// loop over all source values
   ret[i] += data[j]*wActual;
   wActual *= wStep;
  }
 }
 return ret;
}

// C-Quelle kopiert und angepasst von rosettacode.org
// (als Functor-Klasse um sich das Herumschleppen von n und step zu sparen)
template<class T>class fft{
 typedef complex<T> cplx;
 int n;	// Anzahl komplexer Zahlen, popcount() muss 1 sein
 int step;
 void internal(cplx*buf,cplx*out) {
  int hstep=step;
  step<<=1;		// fr die nchste Rekursion verdoppeln
  if (step<n) {
   internal(out, buf);	// die Puffer wechseln jedesmal!
   internal(out+hstep, buf+hstep);
  }
  for (int i = 0; i < n; i += step) {
   cplx t(1,T(-0.5)*i/n,turns);
   t*=out[i+hstep];
   buf[ i   >>1] = out[i]+t;
   buf[(i+n)>>1] = out[i]-t;
  }
  step=hstep;
 }
public:
 fft(cplx*data,int nn):n(nn),step(1) {
  int bytelen=nn*sizeof(cplx);
  cplx*out=(cplx*)_alloca(bytelen);
  memcpy(out,data,bytelen);
  internal(data,out);
 }
};

// Normale FFT mit komplexen Eingabedaten
template<class T> IqData<T> FFT(const IqData<T>&data) {
 IqData<T> ret;
 ret = data;	// deep copy, preserve data
 fft<T>(ret.data(),int(ret.size()));
 return ret;
}

// Dasselbe mit rein reellen Eingabedaten
// (ohne anschlieende Faltung und Lngenhalbierung des Ergebnisses)
template<class T> IqData<T> FFT(const vector<T>&data) {
 IqData<T> ret;
 ret = data;	// take as I data
 fft<T>(ret.data(),int(ret.size()));
 return ret;
}

// Explizite Instantiierungen
template float radians2angle(float,angle_t);
template float radians_norm(float);
template class complex<float>;
template class IqData<float>;
template IqData<float> DFT<float>(const IqData<float>&);	// kopiert
template IqData<float> FFT<float>(const IqData<float>&);	// kopiert auch
template IqData<float> FFT<float>(const vector<float>&);
// Das generiert die Float-Version in der Objektdatei, sonst gibt's einen Linkerfehler
