#pragma once
#include "../util/vector.h"	// Meine vector-Implementierung
#include "../util/util.h"	// my::PI

#define PI_2 (PI/2)
#define PI2 (PI*2)

/********************
 * Ersatz fr numpy *
 * heha 240305	    *
 ********************/

// Ein Vektor mit dem man rechnen kann. Hauptsache T untersttzt die gngigen Operatoren
template<class T> class numvec:public vector<T> {
public:
 typedef T value_type;
private:
// Die Operatoren mssen (offenbar) in Wrapperfunktionen gekapselt werden.
// Denn fr die Operatoren der Standardtypen gibt es keine Funktionszeiger,
// etwa dass ginge "int(*mul_int)(int,int) = int::operator*;"
 static T add     (T a,T b) {return a+b;}
 static T subtract(T a,T b) {return a-b;}
 static T multiply(T a,T b) {return a*b;}
 static T divide  (T a,T b) {return a/b;}
 static T modulo  (T a,T b) {return a%b;}
 static T bitand  (T a,T b) {return a&b;}
 static T bitor   (T a,T b) {return a|b;}
 static T bitxor  (T a,T b) {return a^b;}
 static T shl     (T a,T b) {return a<<b;}
 static T shr     (T a,T b) {return a>>b;}
 static T pos     (T a)     {return +a;}
 static T neg     (T a)     {return -a;}
 static bool land (T a,T b) {return a&&b;}
 static bool lor  (T a,T b) {return a||b;}
 static bool lt   (T a,T b) {return a<b;}
 static bool le   (T a,T b) {return a<=b;}
 static bool gt   (T a,T b) {return a>b;}
 static bool ge   (T a,T b) {return a>=b;}
 static bool ne   (T a,T b) {return a!=b;}
 static bool eq   (T a,T b) {return a==b;}

 numvec<T> produce(T(*cb)(T,T),T o) const{
  numvec<T> r(size());
  iterator d=r._begin, e=r._end;
  for (const_iterator my=_begin; d!=e; d++,my++) *d = cb(*my,o);
  return r;
 }

 numvec<T> produce(T(*cb)(T,T),const vector<T>&o) const{
  numvec<T> r(size());
  vector<T>::const_iterator b=o.begin(), s=b, e=o.end();
  const_iterator my=_begin;
  if (b!=e) for (iterator d=r._begin; d!=r._end; d++,my++) {
   *d = cb(*my,*s);
   if (++s==e) s=b;
  }
  return r;
 }

public:
 numvec(size_t l):vector<T>(l) {}
 numvec<T> operator+(T o)		const {return produce(add,o);}
 numvec<T> operator+(const vector<T>&o)	const {return produce(add,o);}
 numvec<T> operator-(T o)		const {return produce(subtract,o);}
 numvec<T> operator-(const vector<T>&o)	const {return produce(subtract,o);}
 numvec<T> operator*(T o)		const {return produce(multiply,o);}
 numvec<T> operator*(const vector<T>&o)	const {return produce(multiply,o);}
 numvec<T> operator/(T o)		const {return produce(divide,o);}
 numvec<T> operator/(const vector<T>&o)	const {return produce(divide,o);}
 numvec<T> operator%(T o)		const {return produce(modulo,o);}
 numvec<T> operator%(const vector<T>&o)	const {return produce(modulo,o);}
};

/* Eine komplexe Zahl. Es soll mit MSVC6 kompatibel bleiben.
 * Dazu Arrays (Vektoren) von Skalaren und komplexen Zahlen.
 * Arrays von Arrays wurden nicht getestet und werden nicht gebraucht.
 * Normalerweise msste der zweite Operand jeweils einen anderen Basistypen aufweisen drfen.
 * Das ist jedoch mit MSVC6 auf keinen Fall hinzukriegen.
 * Versuche dazu fhrten u.a. zu Inkompatibilitten mit MSVC2008.
 */

enum angle_t{	// A period is:
 radians,	// 2*PI
 turns,		// 1
 degrees=360,	// 360
 gradians=400,	// 400
};

template<class T> T radians2angle(T,angle_t);
template<class T> T radians2degrees(T x) {return radians2angle<T>(x,degrees);}
template<class T> T radians_norm(T);	// brings T into range [-PI..+PI>

template<class T> class complex {
public:
 T x,y;			// Im Gegensatz zu std::complex sind die Elemente zugreifbar
 typedef T value_type;
 T real() const {return x;}
 T imag() const {return y;}
// Ntzliche Funktionen, NICHT in std::complex
 T abs() const;			// erfordert math.h
 T arg(angle_t=radians) const;	// erfordert math.h
 complex& set(T re, T im = 0) {x=re; y=im; return*this;}
 complex& set(T abs, T arg, angle_t ang);
// Konstruktoren
 complex(T re = 0, T im = 0):x(re),y(im) {}	// Defaultkonstruktor initialisiert mit 0
 complex(T abs, T arg, angle_t ang) {set(abs,arg,ang);}	// Winkelargument-Konstruktor
// complex(const complex& o):x(o.x),y(o.y) {}	// = default, also ein Kopierkonstruktor
// Zuweisungs-Operatoren
// complex<T>& operator =(const complex<T>& o)	{x = o.x; y = o.y; return*this;}
 complex<T>& operator+=(const complex<T>& o)	{x+= o.x; y+= o.y; return*this;}
 complex<T>& operator-=(const complex<T>& o)	{x-= o.x; y-= o.y; return*this;}
 complex<T>& operator*=(const complex<T>& o)	{return *this = operator*(o);}
 complex<T>& operator/=(const complex<T>& o)	{return *this = operator/(o);}
 complex<T>& operator =(T o)			{x = o; y = 0;	return*this;}	// nur Realteil
 complex<T>& operator+=(T o)			{x+= o;		return*this;}
 complex<T>& operator-=(T o)			{x-= o;		return*this;}
 complex<T>& operator*=(T o)			{x*= o; y*= o;	return*this;}	// skalieren
 complex<T>& operator/=(T o)			{x/= o; y/= o;	return*this;}
// Infix-Operatoren, generieren neues Objekt (Platz wird vom Aufrufer per __$ReturnUdt bereitgestellt)
 complex<T> operator+(const complex<T>& o) const {return complex<T>(x+o.x,y+o.y);}
 complex<T> operator+(T o)		   const {return complex<T>(x+o,y);}
 complex<T> operator-(const complex<T>& o) const {return complex<T>(x-o.x,y-o.y);}
 complex<T> operator-(T o)		   const {return complex<T>(x-o,y);}
 complex<T> operator*(const complex<T>& o) const;	// kompliziert, nicht inline
 complex<T> operator*(T o)		   const {return complex<T>(x*o,y*o);}
 complex<T> operator/(const complex<T>& o) const;	// kompliziert, nicht inline
 complex<T> operator/(T o)		   const {return complex<T>(x/o,y/o);}
// Infix-Operatoren, die einfache Datentypen liefern
 bool operator==(const complex<T>& o) const {return x==o.x && y==o.y;}
 bool operator!=(const complex<T>& o) const {return !operator==(o);}
// Unre Prfix-Operatoren, generieren neues Objekt
 complex<T> operator-() const {return complex<T>(-x,-y);}	// unres Minus = Vorzeichenumkehr
 bool operator!() const {return !isNotZero();}			// Bang = isZero()
 complex<T> operator~() const {return complex<T>(x,-y);}	// Tilde = konjugiert komplex
// Sonstiges
 bool isNotZero() const {return x||y;}
 complex<T> operator^(char) const;	// wenn ich mal viel Zeit habe; Exponent -1 = Kehrwert
};

// freestanding operation if complex operator is on right side
template<class T> complex<T> operator/(const T&t, const complex<T>&o) {complex<T>r(t,0); return r/=o;}

/* Ein Array variabler Lnge aus komplexen Zahlen.
   Baut in hohem Mae auf complex<T> auf. */

template<class T=double>class IqData:public vector<complex<T> > {
public:
 typedef complex<T> value_type;
private:
// Eine Funktion die mithilfe eines Memberfunktionszeigers
// ein neues Array mit den angegebenen Funktionswerten generiert
 numvec<T> extract(T(value_type::*cb)()const) const{
  numvec<T>ret(size());
  for (size_t i=0; i<size(); i++) ret[i] = (operator[](i).*cb)();
  return ret;
 }
// Dito mit Winkeltyp-bergabe
 numvec<T> extract(T(value_type::*cb)(angle_t)const,angle_t ang) const{
  numvec<T>ret(size());
  for (size_t i=0; i<size(); i++) ret[i] = (operator[](i).*cb)(ang);
  return ret;
 }
// Vier forEach-Routinen fr vier Quelltypen: float, vector<float>, complex, vector<complex>

 IqData<T>& forEach(value_type&(value_type::*cb)(const complex<T>&),const complex<T>&o) {
  for (iterator d=_begin; d!=_end; d++) (d->*cb)(o);
  return*this;
 }
 IqData<T>& forEach(value_type&(value_type::*cb)(T),T o) {
  for (iterator d=_begin; d!=_end; d++) (d->*cb)(o);
  return*this;
 }
 IqData<T>& forEach(value_type&(value_type::*cb)(const complex<T>&),const IqData<T>&o) {
  IqData<T>::const_iterator b=o._begin, s=b, e=o._end;
  if (b!=e) for (iterator d=_begin; d!=_end; d++) {
   (d->*cb)(*s);
   if (++s==e) s=b;	// Bei krzerem zweiten Vektor wird dieser vom Anfang zyklisch verarbeitet!
  }
  return*this;
 }
 IqData<T>& forEach(value_type&(value_type::*cb)(T),const vector<T>&o) {
  vector<T>::const_iterator b=o.begin(), s=b, e=o.end();
  if (b!=e) for (iterator d=_begin; d!=_end; d++) {
   (d->*cb)(*s);
   if (++s==e) s=b;
  }
  return*this;
 }

// Vier produce-Routinen fr vier Quelltypen: float, vector<float>, complex, vector<complex>
 IqData<T> produce(value_type (value_type::*cb)(const complex<T>&) const,const value_type&o) const{
  IqData<T> r(size());
  iterator d=r._begin, e=r._end;
  for (const_iterator my=_begin; d!=e; d++,my++) *d = (my->*cb)(o);
  return r;
 }
 IqData<T> produce(value_type (value_type::*cb)(T) const,T o) const{
  IqData<T> r(size());
  iterator d=r._begin, e=r._end;
  for (const_iterator my=_begin; d!=e; d++,my++) *d = (my->*cb)(o);
  return r;
 }
 IqData<T> produce(value_type (value_type::*cb)(char) const,char o) const{
  IqData<T> r(size());	// for operator^ only
  iterator d=r._begin, e=r._end;
  for (const_iterator my=_begin; d!=e; d++,my++) *d = (my->*cb)(o);
  return r;
 }
 IqData<T> produce(value_type (value_type::*cb)(const complex<T>&) const,const IqData<T>&o) const{
  IqData<T> r(size());
  const_iterator b=o._begin, s=b, e=o._end, my=_begin;
  if (b!=e) for (iterator d=r._begin; d!=r._end; d++,my++) {
   *d = (my->*cb)(*s);
   if (++s==e) s=b;
  }
  return r;
 }
 IqData<T> produce(value_type (value_type::*cb)(T) const,const vector<T>&o) const{
  IqData<T> r(size());
  vector<T>::const_iterator b=o.begin(), s=b, e=o.end();
  const_iterator my=_begin;
  if (b!=e) for (iterator d=r._begin; d!=r._end; d++,my++) {
   *d = (my->*cb)(*s);
   if (++s==e) s=b;
  }
  return r;
 }
 IqData<T> produce(value_type (value_type::*cb)() const) const{
  IqData<T> r(size());
  const_iterator my=_begin;
  for (iterator d=r._begin; d!=r._end; d++,my++) *d = (my->*cb)();
  return r;
 }

// ein Callback fr any() und all()
 bool any(bool (value_type::*cb)() const) const{
  for (const_iterator my=_begin; my!=_end; my++) if ((my->*cb)()) return true;
  return false;
 }
 bool all(bool (value_type::*cb)() const) const{
  for (const_iterator my=_begin; my!=_end; my++) if (!(my->*cb)()) return false;
  return true;
 }

public:
 IqData():vector<value_type>() {}
 explicit IqData(size_t n):vector<value_type>(n) {}	// Das zugrunde liegende Array wird schon mal alloziert, die Gre festgelegt.
// Da der Defaultkonstruktor von complex<T> nullsetzt, ist das Array mit Nullen gefllt.
// Der Standard-Kopierkonstruktor macht eine flache Kopie.
// Der Kopierkonstruktor von std::vector macht eine tiefe Kopie.
// Es ist mir nicht klar was ein von vector<> vererbter Konstruktor tut.
// IqData(const IqData<T>&o);
// Da keine Move-Semantik zur Verfgung steht und so "return array" zu elenden Kopieraktionen ausarten
// lasse ich den Kopierkonstruktor flach und hoffe auf gutes Gelingen.

// Extraktoren, liefern neues vector<T>-Objekt. Ohne nervigen Exceptionhandler.
 numvec<T>real() const {return extract(&value_type::real);}
 numvec<T>imag() const {return extract(&value_type::imag);}
 numvec<T>abs() const {return extract(&value_type::abs);}
 numvec<T>arg(angle_t ang) const {return extract(&value_type::arg,ang);}
// Zuweisungs-Operatoren
// IqData<T>&operator =(const IqData<T>&o) {
//	if (this==&o) return;
//	commit_undefined(o.size());
//	memcpy(data(),o.data(),size()*sizeof(value_type));
//	return*this;
// }	// tiefe Kopie
// <O> should always be a scalar or complex type, not a vector type
 IqData<T>&operator =(const vector<T>&o) {commit_undefined(o.size()); return forEach(&value_type::operator =,o);}
// Bei unterschiedlichen Arraylngen bleibt die Lnge des Ziel-Arrays bestehen(!),
// und das Quell-Array wird, wenn krzer, zyklisch auf das Ziel-Array angewendet.
 IqData<T>&operator+=(const value_type&o)	{return forEach(&value_type::operator+=,o);}
 IqData<T>&operator+=(T o)			{return forEach(&value_type::operator+=,o);}
 IqData<T>&operator+=(const IqData<T>&o)	{return forEach(&value_type::operator+=,o);}
 IqData<T>&operator+=(const vector<T>&o)	{return forEach(&value_type::operator+=,o);}
 IqData<T>&operator-=(const value_type&o)	{return forEach(&value_type::operator-=,o);}
 IqData<T>&operator-=(T o)			{return forEach(&value_type::operator-=,o);}
 IqData<T>&operator-=(const IqData<T>&o)	{return forEach(&value_type::operator-=,o);}
 IqData<T>&operator-=(const vector<T>&o)	{return forEach(&value_type::operator-=,o);}
 IqData<T>&operator*=(const value_type&o)	{return forEach(&value_type::operator*=,o);}
 IqData<T>&operator*=(T o)			{return forEach(&value_type::operator*=,o);}
 IqData<T>&operator*=(const IqData<T>&o)	{return forEach(&value_type::operator*=,o);}
 IqData<T>&operator*=(const vector<T>&o)	{return forEach(&value_type::operator*=,o);}
 IqData<T>&operator/=(const value_type&o)	{return forEach(&value_type::operator/=,o);}
 IqData<T>&operator/=(T o)			{return forEach(&value_type::operator/=,o);}
 IqData<T>&operator/=(const IqData<T>&o)	{return forEach(&value_type::operator/=,o);}
 IqData<T>&operator/=(const vector<T>&o)	{return forEach(&value_type::operator/=,o);}
// Infix-Operatoren, generieren neues Objekt
// Bei unterschiedlichen Arraylngen diktiert der linke Operand die Lnge des Ergebnis-Arrays,
// und der rechte Operand wird, wenn krzer, zyklisch auf das Ergebnis angewendet.
 IqData<T>operator+(const value_type&o) const	{return produce(&value_type::operator+,o);}
 IqData<T>operator+(T o)		const	{return produce(&value_type::operator+,o);}
 IqData<T>operator+(const IqData<T>&o)	const	{return produce(&value_type::operator+,o);}
 IqData<T>operator+(const vector<T>&o)	const	{return produce(&value_type::operator+,o);}
 IqData<T>operator-(const value_type&o) const	{return produce(&value_type::operator-,o);}
 IqData<T>operator-(T o)		const	{return produce(&value_type::operator-,o);}
 IqData<T>operator-(const IqData<T>&o)	const	{return produce(&value_type::operator-,o);}
 IqData<T>operator-(const vector<T>&o)	const	{return produce(&value_type::operator-,o);}
 IqData<T>operator*(const value_type&o) const	{return produce(&value_type::operator*,o);}
 IqData<T>operator*(T o)		const	{return produce(&value_type::operator*,o);}
 IqData<T>operator*(const IqData<T>&o)	const	{return produce(&value_type::operator*,o);}
 IqData<T>operator*(const vector<T>&o)	const	{return produce(&value_type::operator*,o);}
 IqData<T>operator/(const value_type&o) const	{return produce(&value_type::operator/,o);}
 IqData<T>operator/(T o)		const	{return produce(&value_type::operator/,o);}
 IqData<T>operator/(const IqData<T>&o)	const	{return produce(&value_type::operator/,o);}
 IqData<T>operator/(const vector<T>&o)	const	{return produce(&value_type::operator/,o);}
// Unre Prfix-Operatoren, generieren neues Objekt
 IqData<T>operator-()			const	{return produce(&value_type::operator-);}	// Vorzeichenumkehr
 IqData<T>operator~()			const	{return produce(&value_type::operator~);}	// Tilde = konjugiert komplex
// Verschiedenes
 bool isNotZero() const {return any(&value_type::isNotZero);}	// false wenn alle Elemente 0 sind
 bool operator!() const	{return !isNotZero();}
 IqData<T>operator^(char o)		const	{return produce(&value_type::operator^,o);}
 T*lineardata() {return &_begin->x;}
 const T*lineardata() const {return &_begin->x;}
};

// Aus Wikipedia
template<class T> IqData<T> DFT(const IqData<T>&);
template<class T> IqData<T> FFT(const IqData<T>&);
template<class T> IqData<T> FFT(const vector<T>&);

// Unklar was das macht
//template<class T> IqData<T> ZoomFFT(const IqData<T>&, T f1, T f2);
