#pragma once
#include "scale.h"
#include "tick.h"
#include "range.h"
#include "../util/vector.h"

struct AXISATTR{
 DWORD	side;
 char	iaxis,		// Index in vector<uniq_ptr<Axis>> pl.axes
	iplotxy;	// Plot-Zeile oder Spalte (je nach side.0)
 union{
  struct{
   char	col,row;	// Spalte und Zeile im GridLayout
  };
  Point<char>cr;	// col und row
 };
 char sub_cr,rsv;	// Nummer im ggf. aufgeteilten Scale-Feld
 BYTE shrink_time,shrink_perc;
 cstr<> name, unit;
 Float precis;
 Font::Data fdtick,fdtitle;
 AXISATTR():side(0),iaxis(-1),iplotxy(-1),col(-1),row(-1),sub_cr(-1),shrink_time(10),shrink_perc(1),precis(0) {}
};


class XAxis;
class YAxis;
class Trace;
class PlotXY;

class Axis:public AXISATTR{
  friend class XAxis;
  friend class YAxis;
  friend class Trace;
  friend class PlotXY;
  friend class DisplayRange;
  friend class WindowTextParser;
 enum {
// Einzelbits:
  bVertical,
  bBottomRight,
  bReversed,
  bLog,
  bTime,		// Sonderformatierung
  bSeparate,		// Gesonderte Plotzeile/Spalte erzwingen (typischerweise digital)
  bUnused,
  bVisible,
  bAutoExpand,	// Skale passt sich (sofort) an Messdaten an
  bAutoShrink,	// Skale passt sich (verzgert) an Messdaten an
  bIntegralB,	// Anfang auf nchsten ganzen Tick abrunden
  bIntegralE,	// Ende auf nchsten ganzen Tick aufrunden
  bShowB,	// Wahren Skalenanfang anzeigen, sonst ohne Beschriftung
  bShowE,	// Wahres Skalenende anzeigen, sonst ohne Beschriftung
  bShowZ,	// Null ist stets in der Skale (nie bei LOG)
  bCenterZ,	// Null in der Mitte (nicht mit LOG oder ABSTIME vereinbar)
  bCommonZ,	// Gemeinsame Nulllinie mit anderen Skalen derselben Plotarea (knifflig ohne ZEROCENTER!!)
//  ABSTIME = 0x18,	// Sonderformatierung
    DIGITAL = 0x11,	// nur Y-Achse: Beide Seiten sowie feste Hhe, immer mit SEPARATE
    INVAL_RANGE = 0x10000,
    INVAL_TICKS = 0x20000,
  // Wichtige Kombinationen:
  BOTTOM=2,	// X-Achse unten von links nach rechts
  LEFT=5,	// Y-Achse links von unten nach oben
  RIGHT=7,	// Y-Achse rechts von unten nach oben
 };
 int shrink_countdown;
//  BYTE ftick_korr;	// schriftgrenabhngiger Korrekturwert von TA_BASELINE auf (fehlendes) TA_MIDDLE (nur Y-Achsen)
  Margin<>padding;	// Die Skale ist im Rechteck (ein wenig) eingerckt je nach Linienbreite und Markergre:
// rechts mindestens 1 weil die Linienbreite mindestens 1 Pixel ist.
// rechts+links die grte Linienbreite oder grte Markerbreite aller Traces
// oben und unten Null.
// Bei vertikalen Achsen entsprechend an der Hauptdiagonalen gestrzt.
// Bei gestrzten Achsen (bei Y der Normalfall) NICHT gestrzt.
  typedef vector<Tick> ticks_t;
  ticks_t ticks;
  typedef vector<MAJORTICK> labels_t;
  labels_t userlabels;	// Vorgabe-Labels
  typedef vector<Trace*> traces_t;	// hier kein uniq_ptr!
  traces_t traces;	// Zugeordnete Kurvenzge
  bool anyTrace(bool (Trace::*testfunc)()const,bool runwhile=false) const;
  bool allTraces(bool (Trace::*testfunc)()const) const {return anyTrace(testfunc,true);}
  Color cfore,ctitl;	// Hintergrundfarbe, Vordergrundfarbe, Titelfarbe
  Pen ptick;		// Stift fr Skalenstriche
  Font ftick,ftitl;	// Schrift fr Achse und Achsentitel
  char scaleInf(Float v,int&V) const;
  enum {
    CINF=127,CNAN=-128		// Ergebnis von scaleInf: Bewertung des Float-Arguments
  };
  Range<Float> dataRange;	// Bereich fr ALLE (sichtbaren) Traces, exklusive Inf und NaN
  DisplayRange dispRange;	// ggf. durch Zoom verndert, sonst dataRange umfassend
  TCHAR tickformat[8];		// maximal "%.6f\0" = 6 TCHARs
  bool unitToTickFormat() const;
  void buildTickFormat();
  int genLinTicks(DWORD flags,int intervals);
  int genLogTicks(DWORD flags,int intervals);
  Axis(PlotXY&,const AXISATTR&);
  void buildLabels();
//  virtual void paint()=0;
//  virtual int needSpace(bool/*,int*/)=0;
  Range<Float> Axis::allTracesRange() const;
  bool operator==(const AXISATTR&a) const;
  bool updateScaling(bool force);
  bool checkAutorange();
  void checkRange(bool allowShrink=false);
  void distributeTicks();
  bool visible() const {return !!(side&1<<bVisible);}
  void setVisible(bool v) {if (v) side|=1<<bVisible; else side&=~(1<<bVisible);}
  bool isDigital() const {return (side&DIGITAL)==DIGITAL;}
  bool isTime() const {return (side&DIGITAL)==1<<bTime;}
  bool isLog() const {return !!(side&1<<bLog);}
//  int getTitle(TCHAR*buf,int buflen) const;
//  int axistext(TCHAR*buf,int blen) const;
//  TCHAR*axistext() const;
  cstr<>title;
  void makeTitle();
  int textWidth(const TCHAR*, int) const;
  void onTimer(int ms) {if (shrink_countdown>0 && (shrink_countdown-=ms)<=0) checkRange(true);}
  Tick*findEditableLabel(const Point<>&pt);	// nullptr when no match, <pt> is local to Axis rectangle (left/top = 0/0)
 virtual void paint(HDC)=0;
 virtual int needSpace(Margin<>&)=0;
 int FloatStr(Float,TCHAR*,int) const;
 Rect<>allTicksRect();	// Umhllende aller Textrechtecke
 void prepareMenu(HMENU sub);
 static INT_PTR _stdcall BereichDlgProc(HWND Wnd, UINT msg, WPARAM wParam, LPARAM lParam);
public:
 PlotXY&pl;
 Scale&scale;
 struct Metrics{
  char ticklen;
  char tickwidth;
  char gap;		// Platz zwischen Text und Strich, zwischen Text und oberen Rand
  char fontheight;
  char lastSpaceNeeded;
  char steppx;	// Maximale Pixelzahl fr minimale Auflsung: 8 (verhindert gefllte Flchen fr mickrige Rauschamplituden)
  char textycorrect;
  Metrics():ticklen(4),tickwidth(1),gap(1),fontheight(14),lastSpaceNeeded(1),steppx(4),textycorrect(4) {}
 }metrics;
 cstr<>FloatStr(Float) const;
 void onRButtonDown(Point<>&);
};

class XAxis:public Axis{
 virtual void paint(HDC);
 virtual int needSpace(Margin<>&);
 friend class PlotXY;
 XAxis(PlotXY&p,const AXISATTR&a):Axis(p,a) {}	// Konstruktoren durchreichen
};

class YAxis:public Axis{
 virtual void paint(HDC);
 virtual int needSpace(Margin<>&);
 friend class PlotXY;
 YAxis(PlotXY&p,const AXISATTR&a):Axis(p,a) {}
};
